//============================================================================
// Bare-Bones Tablet Interface for C++
//
//  Implementation file
//  This file _should_ contain only common platform-independent code.
//  The actual work is delegated to platform-specific files.
//
//  Author:        William Baxter (baxter@cs.unc.edu)
//  Last Modified: Dec 2005
//============================================================================
//============================================================================
// MIT License:
// Copyright (c) 2002--2009 William Baxter
// 
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
// BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
// ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//============================================================================

#if defined (_WIN32)
#include "bbtabwin.h"
#else
  #error "Platform Not Supported"
#endif


#include "bbtablet.h"

#include <algorithm>

#if defined(DEBUG) || defined(_DEBUG)
static void DebugMsg(const char *fmt, ...)
{
  va_list args;
  va_start(args, fmt);
  vfprintf(stderr, fmt, args);
  va_end(args);
}
#else 
inline void DebugMsg(const char *fmt, ...) { }
#endif




//----------------------------------------------------------------------------

bbTablet &bbTablet::getInstance()
{
  static bbTablet td_singleton;
  return td_singleton;
}

bbTablet::bbTablet()
{
  m_data = new _bbTabletInternal();
}
bbTablet::~bbTablet()
{
  delete m_data;
}


//================== _bbTabletInternalBase ===================================
bool _bbTabletInternalBase::_addObserver(bbTabletObserver& obs)
{
  ObserverList::iterator it=m_observers.begin(), end=m_observers.end();
  if (std::find(it,end,&obs) != end) {
    // Already in the list!  Gack!
    return false;
  }
  m_observers.push_front(&obs);
  return true;
}

bool _bbTabletInternalBase::_removeObserver(bbTabletObserver& obs)
{
  ObserverList::iterator it=m_observers.begin(), end=m_observers.end();
  ObserverList::iterator found = std::find(it,end,&obs);
  if (found == end) {
    // Not in the list!  Gack!
    return false;
  }
  m_observers.erase(found);
  return true;
}

void _bbTabletInternalBase::_callObservers(const bbTabletEvent& evt)
{
  ObserverList::iterator it=m_observers.begin(), end=m_observers.end();
  for (; it!=end; ++it) {
    int ret = (*it)->OnTabletEvent(evt);
  }
}


//================== bbTabletEvent ===========================================
const char *bbTabletEvent::getCursorName() const
{
  return bbTablet::getInstance().getCursorName(type);
}

int bbTabletEvent::getCursorNumButtons() const
{
  return bbTablet::getInstance().getCursorNumButtons(type);
}

const char *bbTabletEvent::getCursorButtonName(int buttonnum) const
{
  return bbTablet::getInstance().getCursorButtonName(type,buttonnum);
}

