//============================================================================
// bbtabwin.h
//
//  Function pointers/prototypes to wintab32 dll functions.
//  Looked up from dll dynamically so that DLL need not be linked with 
//  explicity.  This way programs will run even if WinTab32.dll is not
//  present (just without tablet support).
//
//
//============================================================================
// (MIT License)
// Copyright (c) 2002--2009 William Baxter
// 
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
// BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
// ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//============================================================================

// Here are function pointer types for those
#ifndef BB_TABWIN_H
#define BB_TABWIN_H

# define WIN32_LEAN_AND_MEAN
#include <windows.h> // just for the windows types
#define NOWTFUNCTIONS
#include <wintab.h>

#include "bbtablet.h"
#include "bbtabletpriv.h"

#include <vector>

typedef UINT (API*PFNWTINFO)(UINT, UINT, LPVOID);
typedef HCTX (API*PFNWTOPEN)(HWND, LPLOGCONTEXT, BOOL);
typedef BOOL (API*PFNWTCLOSE)(HCTX);
typedef BOOL (API*PFNWTGET)(HCTX, LPLOGCONTEXT);
typedef BOOL (API*PFNWTSET)(HCTX, LPLOGCONTEXT);
typedef BOOL (API*PFNWTPACKET)(HCTX, UINT, LPVOID);
typedef int  (API*PFNWTPACKETSGET)(HCTX, int, LPVOID);
typedef BOOL (API*PFNWTQUEUEPACKETSEX)(HCTX, UINT FAR *, UINT FAR *);
typedef BOOL (API*PFNWTOVERLAP)(HCTX, BOOL);

extern PFNWTINFO pfWTInfo;
extern PFNWTOPEN pfWTOpen;
extern PFNWTCLOSE pfWTClose;
extern PFNWTGET  pfWTGet;
extern PFNWTSET  pfWTSet;
extern PFNWTPACKET pfWTPacket;
extern PFNWTPACKETSGET pfWTPacketsGet;
extern PFNWTQUEUEPACKETSEX pfWTQueuePacketsEx;
extern PFNWTOVERLAP pfWTOverlap;

enum { BBT_MAX_DEVICES = 10 };

struct _bbTabletInfo
{
  _bbTabletInfo() : tabCtx(0) {}
  HCTX tabCtx;
  float AspectRatio;
  float scaleX;
  float scaleY;
  float scaleZ;
  float sysScaleX;
  float sysScaleY;
  float scalePress;
  float scaleTPress;
  float scaleAng;
};

class _bbTabletInternal : public _bbTabletInternalBase
{
public:
  _bbTabletInternal();
  ~_bbTabletInternal()
  {
    _removeCallbackHooks();
  }

  HWND m_hwnd;
  _bbTabletInfo m_dev[BBT_MAX_DEVICES];
  int m_nextQueryDev;
  std::vector<_bbTabletCursor> m_cursors;

  void _convertOrientation2AngleAxis(
    const ORIENTATION &orient, float scale, float *angle, float axis[3]);
  bool _reopenDeviceContext(int devID);
  bool _setDevicePointerMode(int mode,int devID);
  int _getDevicePointerMode(int devID);

  // CallBack Setup
  static UINT  ms_msgBase;
  static HHOOK ms_msgFilterHook;

  bbTabletEvent m_evt; // for sending events to callbacks
  static LRESULT CALLBACK msgProcFilter(int nCode, WPARAM wParam, LPARAM lParam);
  void _setupCallbackHooks();
  void _removeCallbackHooks();
  void msgProcProcess(const MSG& msg);
  void msgProcProcessWTPacket(HCTX hCtx);
  void _setMessagesEnabled(bool trueFalse);

  friend class bbTabletEvent;
};



#endif /* BB_TABWIN_H */
