/*
 *  LICENSE:
 *  I place this sample code in the PUBLIC DOMAIN
 */

#include <GL/glut.h>
#include "bbtablet.h"
#include "gltext.hpp"
//#include <hash_map>
//using std::hash_map;
#include <map>
using std::map;

const float rad2deg = 180.0f/3.14159265;
float g_angle = 0;
float g_axis[3]={0,1,0};
float g_pos[3]={0,0,0};
inline float rand_range(float lo, float hi)
{
  return ( rand()/float(RAND_MAX) * (hi-lo) ) + lo;
}

struct StylusData
{
  unsigned int id;
  float axis[3];
  float angle;
  float x,y,z;
  float pressure;
  long lastTimestamp;
  float color[3];
  bool inverted;
};

//typedef hash_map<unsigned int, StylusData> StylusDataMap;
typedef map<unsigned int, StylusData> StylusDataMap;
StylusDataMap g_stylusData;
bbTabletEvent g_evt;
bool g_showStatus = false;
char g_statusMsg[1024]="hi";


void processTabletEvents()
{
  bbTabletDevice &td = bbTabletDevice::getInstance();

  while (td.getNextEvent(g_evt)) {
    g_evt.getOrientation(g_angle, g_axis);
    g_angle *= rad2deg;
    g_pos[0] = g_evt.x * 20.0f - 10.0f;
    g_pos[1] = g_evt.z * 20 ;
    g_pos[2] = -(g_evt.y* 20.0f - 10.0f);


    StylusDataMap::iterator idata = g_stylusData.find(g_evt.id);
    if (idata == g_stylusData.end()) {
      StylusData d;
      d.color[0]=rand_range(.3,.9);
      d.color[1]=rand_range(.3,.9);
      d.color[2]=rand_range(.3,.9);
      g_stylusData[g_evt.id]=d;
      idata = g_stylusData.find(g_evt.id);
    }
    StylusData &d = idata->second;
    d.x = g_pos[0];
    d.y = g_pos[1];
    d.z = g_pos[2];
    d.inverted = g_evt.inverted();
    d.angle = g_angle;
    d.pressure = g_evt.pressure;
    memcpy(d.axis,g_axis,3*sizeof(g_axis[0]));
  }
}

#ifdef _WIN32
#include <windows.h>
HWND getCurrentHWND() 
{
  // get current DC from wgl
  HDC gldc = wglGetCurrentDC();
  // obtain the HWND from that
  HWND win = WindowFromDC(gldc);
  return win;
}
#endif

inline void setAmbient(float r, float g, float b) {
  float v[4]={r,g,b,1};
  glMaterialfv(GL_FRONT,GL_AMBIENT,v);
}
inline void setDiffuse(float r, float g, float b) {
  float v[4]={r,g,b,1};
  glMaterialfv(GL_FRONT,GL_DIFFUSE,v);
}
inline void setAmbient(const float c[3]) {
  float v[4]={c[0],c[1],c[2],1};
  glMaterialfv(GL_FRONT,GL_AMBIENT,v);
}
inline void setDiffuse(const float c[3]) {
  float v[4]={c[0],c[1],c[2],1};
  glMaterialfv(GL_FRONT,GL_DIFFUSE,v);
}
inline void setSpecular(float r, float g, float b) {
  float v[4]={r,g,b,1};
  glMaterialfv(GL_FRONT,GL_SPECULAR,v);
}
inline void setExponent(float n) {
  glMaterialf(GL_FRONT,GL_SHININESS,n);
}

void placeLights() 
{
  float lpos0[] = { -5,10,5,1 };
  glLightfv(GL_LIGHT0, GL_POSITION, lpos0 );
  float lpos1[] = { 3,-3,3,0 };
  glLightfv(GL_LIGHT1, GL_POSITION, lpos1 );

  // Hmm these spotlights don't seem to be doing anything...
  float lpos2[] = { -0,3,-0,1 };
  glLightfv(GL_LIGHT2, GL_POSITION, lpos2 );
  float lpos3[] = {  6,10,-6,1 };
  glLightfv(GL_LIGHT3, GL_POSITION, lpos3 );

  float sdir[3] = {0, -1, 0};
  glLightfv(GL_LIGHT2, GL_SPOT_DIRECTION, sdir );
  glLightfv(GL_LIGHT3, GL_SPOT_DIRECTION, sdir );
  glLightf(GL_LIGHT2, GL_SPOT_EXPONENT, 15.0f);
  glLightf(GL_LIGHT3, GL_SPOT_EXPONENT, 15.0f);
  glLightf(GL_LIGHT2, GL_SPOT_CUTOFF, 20 );
  glLightf(GL_LIGHT3, GL_SPOT_CUTOFF, 20 );
}

void drawGround()
{
  setDiffuse(.6,.7,.5);
  setAmbient(.5,.5,.5);
  glNormal3f(0,1,0);
  int nW = 15;
  int nH = 15;
  for (int i=0; i<nW; i++) {
    float x0 = float(i)*20/float(nW)-10;
    float x1 = float(i+1)*20/float(nW)-10;
    glBegin(GL_QUAD_STRIP);
    for (int j=0; j<=nH; j++) {
      float y = float(j)*20/float(nH)-10;
      glVertex3f(x0, 0, y);
      glVertex3f(x1, 0, y);
    }
    glEnd();
  }
}

void drawStylus(const StylusData &d)
{
  static GLUquadric*barrel=0;
  if (!barrel)
    barrel = gluNewQuadric();

  setDiffuse(d.color);
  setAmbient(d.color);

  bool eraserMode = d.inverted;

  float tipLength = 2.0-d.pressure;
  float tipWidth = 0.6+0.2*d.pressure;
  float barrelLength = 6.0;
  float eraserRadius = 0.8;

  glPushMatrix();
  glTranslatef(d.x,d.y,d.z);
  glRotatef(d.angle,d.axis[0],d.axis[2],-d.axis[1]);
  if (eraserMode) {
    glTranslatef(0,-(barrelLength+eraserRadius),0);
  } else {
    glTranslatef(0,tipLength,0);
  }
  glRotatef(90,1,0,0);
  {
    // Ball/eraser
    glTranslatef(0,0,-6);
    glutSolidSphere(eraserRadius,16,16);
    glTranslatef(0,0,6);

    // Barrel
    glTranslatef(0,0,-6);
    gluCylinder( barrel, 0.6, tipWidth, barrelLength, 8, 10);
    glTranslatef(0,0,6);

    // Tip
    glutSolidCone(tipWidth,tipLength,8,8);

    // Button
    float t=0.5;
    setDiffuse((1-t)*d.color[0]+t*1, (1-t)*d.color[1]+t*1, (1-t)*d.color[2]+t*1);
    //setAmbient(d.color);
    glTranslatef(0,0.5,-1.5);
    glutSolidCube(0.5);
    glTranslatef(0,-0.5,+1.5);
  }

  glPopMatrix();
}

void drawStylii()
{
  StylusDataMap::const_iterator st=g_stylusData.begin(),
    stend=g_stylusData.end();
  for (; st!=stend; ++st) 
  {
    drawStylus(st->second);
  }
}

void drawStats()
{
  if (!g_showStatus) return;
  float angle, axis[3];
  g_evt.getOrientation(angle, axis);
  sprintf(g_statusMsg,
    "\
id = %d\n\
cursor type = %d\n\
cursor name = %s\n\
button mask = %x\n\
position = (%0.3f, %0.3f, %0.3f)\n\
pressure = %0.3f\n\
tangent pressure = %0.3f\n\
angle,axis = %0.3f, (%0.3f,%0.3f,%0.3f)\n\
RollPitchYaw = (%ld, %ld, %ld)\n\
",
    g_evt.id,
    g_evt.type,
    g_evt.getCursorName(),
    g_evt.buttons,
    g_evt.x,
    g_evt.y,
    g_evt.z,
    g_evt.pressure,
    g_evt.tpressure,
    angle,
    axis[0],axis[1],axis[2],
    g_evt.getRawRoll(),
    g_evt.getRawPitch(),
    g_evt.getRawYaw()
    );

  Text(10,glutGet(GLUT_WINDOW_HEIGHT)-GetFontHeight()-10,g_statusMsg);
}

void idleFunc()
{
  processTabletEvents();
  glutPostRedisplay();
}

void displayFunc()
{
  glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);

  placeLights();

  drawGround();

  drawStylii();

  drawStats();

  glutSwapBuffers();
}

void keyboardFunc(unsigned char key, int x, int y)
{
  if (key==' ') {
    g_showStatus = !g_showStatus;
  }
  else if (key==27) {
    exit(0);
  }
}

void specialFunc(int key, int x, int y)
{
}

void mouseFunc(int button, int state, int x, int y)
{

}

void motionFunc(int x, int y)
{

}

void reshapeFunc(int W, int H) 
{
  glViewport(0, 0, W, H);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45,W/double(H),.1,100);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(
    0,5,20, // eye
    0,2,0,  // center
    0,1,0); // up

  static bool firstTime=true;
  if (firstTime) {
    void *wnd = getCurrentHWND();
    bbTabletDevice &td = bbTabletDevice::getInstance( );
    td.initTablet( wnd, bbTabletDevice::SEPARATE_POINTER );
  }

}

void placeWindow(int W, int H)
{
  int scrnW = glutGet(GLUT_SCREEN_WIDTH);
  int scrnH = glutGet(GLUT_SCREEN_HEIGHT);
  if (scrnW<W) W=scrnW;
  if (scrnH<H) H=scrnH;
  glutInitWindowPosition((scrnW-W)/2, (scrnH-H)/2);
  glutInitWindowSize(W,H);
}

int main(int argc, char* argv[])
{
  glutInit(&argc,argv);

  glutInitDisplayMode(GLUT_DOUBLE|GLUT_DEPTH);

  placeWindow(640,480);
  int winid = glutCreateWindow("bbTablet Visual Test");

  glutIdleFunc(idleFunc);
  glutDisplayFunc(displayFunc);
  glutKeyboardFunc(keyboardFunc);
  glutSpecialFunc(specialFunc);
  glutReshapeFunc(reshapeFunc);

  glClearColor(.2,.3,.3,1);


  glEnable(GL_DEPTH_TEST);
  //glDisable(GL_COLOR_MATERIAL);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glEnable(GL_LIGHT2);
  glEnable(GL_LIGHT3);
  float lcolor[4] = { .7,.6,.6,1 };
  glLightfv(GL_LIGHT1, GL_AMBIENT_AND_DIFFUSE, lcolor );
  lcolor[0]=lcolor[1]=lcolor[2]=1.0f;
  glLightfv(GL_LIGHT2, GL_AMBIENT_AND_DIFFUSE, lcolor );
  glLightfv(GL_LIGHT3, GL_AMBIENT_AND_DIFFUSE, lcolor );

  printf("Help:\n\tHit space bar to show stats\n");

  glutMainLoop();

  return 0;
}
