//============================================================================
// texfont.cpp - Implementation File
//
// Description: 
//   <TODO>
//
// Author:  William V. Baxter III
// Created: 16 Dec 2005
//============================================================================

#include "texfont.h"
#include <SDL.h>
#ifdef _WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#include <sstream>
#include <stdarg.h>

static SDL_Surface *processImage(SDL_Surface *image)
{
  Uint8 *rowhi, *rowlo;
  Uint8 *tmpbuf, tmpch;
  int i, j;
  /* GL surfaces are upside-down and RGB, not BGR :-) */
  tmpbuf = (Uint8 *)malloc(image->pitch);
  if ( tmpbuf == NULL ) {
    fprintf(stderr, "Out of memory\n");
    return(NULL);
  }
  rowhi = (Uint8 *)image->pixels;
  rowlo = rowhi + (image->h * image->pitch) - image->pitch;
  for ( i=0; i<image->h/2; ++i ) {
    for ( j=0; j<image->w; ++j ) {
      tmpch = rowhi[j*3];
      rowhi[j*3] = rowhi[j*3+2];
      rowhi[j*3+2] = tmpch;
      tmpch = rowlo[j*3];
      rowlo[j*3] = rowlo[j*3+2];
      rowlo[j*3+2] = tmpch;
    }
    memcpy(tmpbuf, rowhi, image->pitch);
    memcpy(rowhi, rowlo, image->pitch);
    memcpy(rowlo, tmpbuf, image->pitch);
    rowhi += image->pitch;
    rowlo -= image->pitch;
  }
  free(tmpbuf);
  return(image);
}

static SDL_Surface *loadBMPFile(char *filename)
{
  SDL_Surface *image;
    
	image = SDL_LoadBMP(filename);
	if ( image == NULL ) {
		fprintf(stderr, "Unable to load %s: %s\n", filename, SDL_GetError());
		return(NULL);
	}
  return processImage(image);
}

static SDL_Surface *loadBMPMem(const unsigned char *mem, int sz)
{
  SDL_Surface *image = SDL_LoadBMP_RW(SDL_RWFromConstMem(mem,sz),1);
  return processImage(image);
}

#if 0
int LoadGLTextures()   // Load Bitmaps And Convert To Textures
{
  int Status=FALSE;							   // Status Indicator
  SDL_Surface *TextureImage[2];			   // Create Storage Space For The Textures
  memset(TextureImage,0,sizeof(void *)*2);		// Set The Pointer To NULL

  if ((TextureImage[0]=LoadBMP("Data17/Font.bmp")) &&
      (TextureImage[1]=LoadBMP("Data17/Bumps.bmp")))
  {
    Status=TRUE;							// Set The Status To TRUE
    glGenTextures(2, &texture[0]);		  // Create Two Texture

    for (loop=0; loop<2; loop++)
    {
      glBindTexture(GL_TEXTURE_2D, texture[loop]);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
      glTexImage2D(GL_TEXTURE_2D, 0, 3, TextureImage[loop]->w, TextureImage[loop]->h, 0, GL_RGB, GL_UNSIGNED_BYTE, TextureImage[loop]->pixels);
    }
  }
  for (loop=0; loop<2; loop++)
  {
    if (TextureImage[loop])							// If Texture Exists
    {
      SDL_FreeSurface(TextureImage[loop]);		// Free The Image Structure
    }
  }
  return Status;								  // Return The Status
}

GLvoid BuildFont(GLvoid)								// Build Our Font Display List
{
  float	cx;											// Holds Our X Character Coord
  float	cy;											// Holds Our Y Character Coord

  base=glGenLists(256);								// Creating 256 Display Lists
  glBindTexture(GL_TEXTURE_2D, texture[0]);			// Select Our Font Texture
  for (loop=0; loop<256; loop++)						// Loop Through All 256 Lists
  {
    cx=(float)(loop%16)/16.0f;						// X Position Of Current Character
    cy=(float)(loop/16)/16.0f;						// Y Position Of Current Character

    glNewList(base+loop,GL_COMPILE);				// Start Building A List
    glBegin(GL_QUADS);							// Use A Quad For Each Character
    glTexCoord2f(cx,1-cy-0.0625f);			// Texture Coord (Bottom Left)
    glVertex2i(0,0);						// Vertex Coord (Bottom Left)
    glTexCoord2f(cx+0.0625f,1-cy-0.0625f);	// Texture Coord (Bottom Right)
    glVertex2i(16,0);						// Vertex Coord (Bottom Right)
    glTexCoord2f(cx+0.0625f,1-cy);			// Texture Coord (Top Right)
    glVertex2i(16,16);						// Vertex Coord (Top Right)
    glTexCoord2f(cx,1-cy);					// Texture Coord (Top Left)
    glVertex2i(0,16);						// Vertex Coord (Top Left)
    glEnd();									// Done Building Our Quad (Character)
    glTranslated(10,0,0);						// Move To The Right Of The Character
    glEndList();									// Done Building The Display List
  }													// Loop Until All 256 Are Built
}

GLvoid KillFont(GLvoid)									// Delete The Font From Memory
{
  glDeleteLists(base,256);							// Delete All 256 Display Lists
}
#endif




// Helper function to read a piece of data from a stream.
template<class T, class S>
void ReadObject(T& to_read, S& in)
{
  in.read(reinterpret_cast<char*>(&to_read), sizeof(T));
}

// This is how glyphs are stored in the file.
struct GlyphBuffer
{
  unsigned char ascii, width;
  unsigned short x, y;
};

GLTextureFont::GLTextureFont()
  : _line_height(0),
    _texture(0),
    _tex_line_height(0)
{

}

GLTextureFont::GLTextureFont(const char* filename)
  : _line_height(0),
    _texture(0),
    _tex_line_height(0)
{
  load(filename);
}


// Load font data from a byte stream
void GLTextureFont::load(std::istream &input)
{
  if (input.fail() || input.get() != 'F' || input.get() != '0')
    throw std::runtime_error("Not a valid font file.");

  // Get the texture size, the number of glyphs and the line height.
  size_t width, height, n_chars;
  ReadObject(width, input);
  ReadObject(height, input);
  ReadObject(_line_height, input);
  ReadObject(n_chars, input);
  _tex_line_height = static_cast<float>(_line_height) / height;

  // Make the glyph table.
  _glyphs = new Glyph[n_chars];
  for (size_t i = 0; i != 256; ++i)
    _table[i] = NULL;

  // Read every glyph, store it in the glyph array and set the right
  // pointer in the table.
  GlyphBuffer buffer;
  for (size_t i = 0; i < n_chars; ++i){
    ReadObject(buffer, input);
    _glyphs[i].tex_x1 = static_cast<float>(buffer.x) / width;
    _glyphs[i].tex_x2 = static_cast<float>(buffer.x + buffer.width) / width;
    _glyphs[i].tex_y1 = static_cast<float>(buffer.y) / height;
    _glyphs[i].advance = buffer.width;

    _table[buffer.ascii] = _glyphs + i;
  }
  // All chars that do not have their own glyph are set to point to
  // the default glyph.
  Glyph* default_glyph = _table[(unsigned char)'\xFF'];
  // We must have the default character (stored under '\xFF')
  if (default_glyph == NULL)
    throw std::runtime_error("Font file contains no default glyph");
  for (size_t i = 0; i != 256; ++i){
    if (_table[i] == NULL)
      _table[i] = default_glyph;
  }

  // Store the actual texture in an array.
  unsigned char* tex_data = new unsigned char[width * height];
  input.read(reinterpret_cast<char*>(tex_data), width * height);

  // Generate an alpha texture with it.
  glGenTextures(1, &_texture);
  glBindTexture(GL_TEXTURE_2D, _texture);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA8, width, height, 0, GL_ALPHA,
               GL_UNSIGNED_BYTE, tex_data);
  // And delete the texture memory block
  delete[] tex_data;
}

// Load font data from an in-memory file
void GLTextureFont::load(const unsigned char* memptr, size_t sz)
{
  if (memptr==0 || sz == 0 || memptr[0]!='F' || memptr[1] != '0')
    throw std::runtime_error("Not a valid font file.");
  std::string memstr((const char*)memptr,sz); 
  std::stringstream input(memstr, std::ios::in);
  load(input);
}

// Load font data from a file
void GLTextureFont::load(const char* filename)
{
  // Open the file and check whether it is any good (a font file
  // starts with "F0")
  std::ifstream input(filename, std::ios::binary);
  load(input);
}

GLTextureFont::~GLTextureFont()
{
  // Release texture object and glyph array.
  glDeleteTextures(1, &_texture);
  delete[] _glyphs;
}

size_t GLTextureFont::lineHeight() const
{
  return _line_height;
}
size_t GLTextureFont::charWidth(unsigned char c) const
{
  return _table[c]->advance;
}

size_t GLTextureFont::stringWidth(const std::string& str) const
{
  size_t total = 0;
  for (size_t i = 0; i != str.size(); ++i)
    total += charWidth(str[i]);
  return total;
}

size_t GLTextureFont::stringWidth(const char *str, size_t len) const
{
  size_t total = 0;
  for (size_t i = 0; i < len; ++i)
    total += charWidth(str[i]);
  return total;
}

size_t GLTextureFont::stringWidth(const char *str) const
{
  size_t total = 0;
  while (*str) 
    total+=charWidth(*str++);
  return total;
}

void GLTextureFont::drawString(float x, float y, const std::string &str) const
{
  drawString(x,y,str.c_str(),str.size());
}

int GLTextureFont::drawGlyph(float &x,float &y, char c) const
{
  Glyph* glyph = _table[c];
  if (c=='\n') {
    x = 0;
    y += _line_height;
    return 1;
  }
  else if (c=='\t') {
    x += 4*glyph->advance;
  }
  else {
    glTexCoord2f(glyph->tex_x1, glyph->tex_y1);
    glVertex2f(x, y);
    glTexCoord2f(glyph->tex_x1, glyph->tex_y1 + _tex_line_height);
    glVertex2f(x, y + _line_height);
    glTexCoord2f(glyph->tex_x2, glyph->tex_y1 + _tex_line_height);
    glVertex2f(x + glyph->advance, y + _line_height);
    glTexCoord2f(glyph->tex_x2, glyph->tex_y1);
    glVertex2f(x + glyph->advance, y);

    x += glyph->advance;
  }
  return 0;
}

void GLTextureFont::drawString(float x, float y, const char *str, size_t len) const
{
  glBindTexture(GL_TEXTURE_2D, _texture);
  float x0 = x;

  // Simply draw quads textured with the current glyph for every
  // character, updating the x position as we go along.
  glBegin(GL_QUADS);
  for (size_t i = 0; i < len; ++i){
    if (drawGlyph(x,y,str[i])) {
      x = x0;
    }
  }
  glEnd();
}

void GLTextureFont::drawString(float x, float y, const char *str) const
{
  glBindTexture(GL_TEXTURE_2D, _texture);
  float x0 = x;
  // Simply draw quads textured with the current glyph for every
  // character, updating the x position as we go along.
  glBegin(GL_QUADS);
  for (size_t i=0; str[i]; ++i) {
    if (drawGlyph(x,y,str[i])) {
      x = x0;
    }
  }
  glEnd();
}

#ifdef _MSC_VER
#define vsnprintf _vsnprintf
#endif
void GLTextureFont::formatString(float x, float y, const char *format, ...) const
{
  va_list args;
  char buffer[1024];
  va_start(args,format);
  int len = vsnprintf(buffer,1024,format,args);
  va_end(args);

  drawString(x,y,buffer,len);
}

void GLTextureFont::pushOrtho2D()
{
  int WW = SDL_GetVideoSurface()->w;
  int WH = SDL_GetVideoSurface()->h;
  glPushAttrib(GL_TEXTURE_BIT|GL_LIGHTING_BIT|GL_ENABLE_BIT|
               GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_VIEWPORT_BIT);

  glEnable(GL_TEXTURE_2D);

  glDisable(GL_LIGHTING);

  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_DITHER);

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE);			// Set The Blending Function For Translucency

  glDisable(GL_DEPTH_TEST);

  glViewport(0, 0, WW, WH);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  gluOrtho2D(0, WW, WH, 0);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
}

void GLTextureFont::popOrtho2D()
{
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glPopAttrib();
}

//--- Emacs setup ---
// Local Variables:
// c-basic-offset: 2
// indent-tabs-mode: nil
// End:

