//============================================================================
// texfont.h - Declaration File
//
// Description: 
//   <TODO>
//
// Author:  William V. Baxter III
// Created: 16 Dec 2005
//============================================================================

#ifndef TEX_FONT_H_
#define TEX_FONT_H_

#include <fstream>

class GLTextureFont
{
public:
  // Construct but don't load any font
  GLTextureFont();
  // Construct, loading font from the specified 'fontmaker' generated file 
  GLTextureFont(const char* file);
  // Construct, loading font from the specified memory buffer
  // (the memory should be in the same format as a font file created by 
  //  fontmaker)
  GLTextureFont(const unsigned char* membuf, size_t size);
  ~GLTextureFont();

  void load(std::istream& input);
  void load(const char*file);
  void load(const unsigned char*membuf, size_t size);

  // The line height is constant
  size_t lineHeight() const;

  // Width of the given char in texels
  size_t charWidth(unsigned char c) const;
  
  // String width from string
  size_t stringWidth(const std::string& str) const;
  
  // String width from null-terminated char buffer
  size_t stringWidth(const char *str) const;
  
  // String width from char buffer and length
  size_t stringWidth(const char *str, size_t len) const;

  // Draw a std::string at a given position.
  void drawString(float x, float y, const std::string& str) const;
  // Draw a null terminated string at a given position.
  void drawString(float x, float y, const char*str) const;
  // Draw a string with the specified length at a given position.
  void drawString(float x, float y, const char*str, size_t len) const;

  // Draw a string with printf-style format string
  void formatString(float x, float y, const char*format, ...) const;

  // Draw a single glyph.  Update x and y with pos for next char.
  // Return the number of lines advanced.
  int drawGlyph(float &x, float &y, char c) const;

  static void pushOrtho2D();
  static void popOrtho2D();
  
private:
  // Information about a glyph. Tex_y2 can be calculated from tex_y1
  // and _tex_line_height (see below). Advance is the width of the
  // glyph in screen space.
  struct Glyph
  {
    float tex_x1, tex_y1, tex_x2;
    size_t advance;
  };
  // An array to store the glyphs.
  Glyph* _glyphs;
  // A table to quickly get the glyph belonging to a character.
  Glyph* _table[256];
  // The line height, in screen space and in texture space, and the
  // OpenGL id of the font texture.
  size_t _line_height, _texture;
  float _tex_line_height;
};



#endif
//--- Emacs setup ---
// Local Variables:
// c-basic-offset: 2
// indent-tabs-mode: nil
// End:

